from os import environ
import logging
from pathlib import Path
from typing import Any, Type, Tuple
from yaml import safe_load
from importlib.metadata import version as importlib_version

from xdg import BaseDirectory
from pydantic import BaseModel
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)


def get_version():
    version = importlib_version("gisaf-backend")
    if version == "0.0.0":
        from subprocess import run

        try:
            v_git_describe = run(
                ["git", "describe", "--always", "--tags", "--dirty"],
                capture_output=True,
                text=True,
                cwd=Path(__file__).parent.parent,
            )
            v_git_describe.check_returncode()
            version = v_git_describe.stdout.strip()
        except Exception:
            logger.warn("Version not set in Python package and cannot determine ")
            version = "?"
    return version


logger = logging.getLogger(__name__)
ENV = environ.get("env", "prod")

config_files = [
    Path(BaseDirectory.xdg_config_home) / "gisaf" / f"{ENV}.yaml",
    Path(BaseDirectory.xdg_config_home) / "gisaf" / f"{ENV}.yml",
    Path(Path.cwd().root) / "etc" / "gisaf" / f"{ENV}.yaml",
    Path(Path.cwd().root) / "etc" / "gisaf" / f"{ENV}.yml",
]


class DashboardHome(BaseModel):
    title: str = "Gisaf - home/dashboards"
    content_file: Path = (
        Path(Path.cwd().root) / "etc" / "gisaf" / "dashboard_home_content.html"
    )
    footer_file: Path = (
        Path(Path.cwd().root) / "etc" / "gisaf" / "dashboard_home_footer.html"
    )


class GisafConfig(BaseModel):
    title: str = "Gisaf"
    windowTitle: str = "Gisaf"
    debugLevel: str = "INFO"
    dashboard_home: DashboardHome = DashboardHome()
    redirect: str = ""
    use_pretty_errors: bool = False


class SpatialSysRef(BaseModel):
    author: str = "AVSM"
    ellps: str = "WGS84"
    k: int = 1
    lat_0: float = 12.01605433
    lon_0: float = 79.80998934
    no_defs: bool = True
    proj: str = "tmerc"
    towgs84: str = "0,0,0,0,0,0,0"
    units: str = "m"
    x_0: float = 370455.630
    y_0: float = 1328608.994


class RawSurvey(BaseModel):
    spatial_sys_ref: SpatialSysRef = SpatialSysRef()
    srid: int = 910001


class Geo(BaseModel):
    raw_survey: RawSurvey = RawSurvey()
    simplify_geom_factor: int = 10000000
    simplify_preserve_topology: bool = False
    srid: int = 4326
    srid_for_proj: int = 32644


# class Flask(BaseModel):
#     secret_key: str
#     debug: int


class MQTT(BaseModel):
    broker: str = "localhost"
    port: int = 1883


class GisafLive(BaseModel):
    hostname: str = "localhost"
    port: int = 80
    scheme: str = "http"
    redis: str = "redis://localhost"
    mqtt: MQTT = MQTT()


class DefaultSurvey(BaseModel):
    surveyor_id: int = 1
    equipment_id: int = 1


class Survey(BaseModel):
    db_schema_raw: str = "raw_survey"
    db_schema: str = "survey"
    default: DefaultSurvey = DefaultSurvey()


class Crypto(BaseModel):
    secret: str = "Gisaf big secret"
    algorithm: str = "HS256"
    expire: float = 21600


class DB(BaseModel):
    # uri: str
    host: str = "localhost"
    port: int = 5432
    user: str = "gisaf"
    db: str = "gisaf"
    password: str = "secret"
    debug: bool = False
    info: bool = True
    pool_size: int = 10
    max_overflow: int = 10
    echo: bool = False

    def get_sqla_url(self):
        return f"postgresql+asyncpg://{self.user}:{self.password}@{self.host}:{self.port}/{self.db}"

    def get_pg_url(self):
        return f"postgresql://{self.user}:{self.password}@{self.host}:{self.port}/{self.db}"


class Log(BaseModel):
    level: str = "WARNING"


class OGCAPILicense(BaseModel):
    name: str = "CC-BY 4.0 license"
    url: str = "https://creativecommons.org/licenses/by/4.0/"


class OGCAPIProvider(BaseModel):
    name: str = "Organization Name"
    url: str = "https://pygeoapi.io"


class OGCAPIServerContact(BaseModel):
    name: str = "Lastname, Firstname"
    position: str = "Position Title"
    address: str = "Mailing Address"
    city: str = "City"
    stateorprovince: str = "Administrative Area"
    postalcode: int = 0
    country: str = "Country"
    email: str = "you@example.org"
    url: str | None = None


class OGCAPIIdentification(BaseModel):
    title: str = "pygeoapi default instance"
    description: str = "pygeoapi provides an API to geospatial data"
    keywords: list[str] = ["geospatial", "data", "api"]
    keywords_type: str = "theme"
    terms_of_service: str = "https://creativecommons.org/licenses/by/4.0/"
    url: str = "http://example.org"


class OGCAPIMetadata(BaseModel):
    identification: OGCAPIIdentification = OGCAPIIdentification()
    license: OGCAPILicense = OGCAPILicense()
    provider: OGCAPIProvider = OGCAPIProvider()
    contact: OGCAPIServerContact = OGCAPIServerContact()


class ServerBind(BaseModel):
    host: str = "0.0.0.0"
    port: int = 5000


class OGCAPIServerMap(BaseModel):
    url: str = "https://maps.wikimedia.org/osm-intl/{z}/{x}/{y}.png"
    attribution: str = (
        """<a href="https://wikimediafoundation.org/wiki/Maps_Terms_of_Use">Wikimedia maps</a> | Map data &copy; <a href="https://openstreetmap.org/copyright">OpenStreetMap contributors</a>"""
    )


class OGCAPIServer(BaseModel):
    bind: ServerBind = ServerBind()
    url: str = "https://example.org/ogcapi"
    mimetype: str = "application/json; charset=UTF-8"
    encoding: str = "utf-8"
    language: str = "en-US"
    pretty_print: bool = False
    limit: int = 1000
    map: OGCAPIServerMap = OGCAPIServerMap()


class OGCAPI(BaseModel):
    base_url: str = "http://example.org/ogcapi"
    bbox: list[float] = [-180, -90, 180, 90]
    log: Log = Log()
    metadata: OGCAPIMetadata = OGCAPIMetadata()
    server: OGCAPIServer = OGCAPIServer()


class TileServer(BaseModel):
    baseDir: Path = Path(BaseDirectory.xdg_data_home) / "gisaf" / "mbtiles_files_dir"
    useRequestUrl: bool = False
    spriteBaseDir: Path = (
        Path(BaseDirectory.xdg_data_home) / "gisaf" / "mbtiles_sprites_dir"
    )
    spriteUrl: str = "/tiles/sprite/sprite"
    spriteBaseUrl: str = "https://gisaf.example.org"
    openMapTilesKey: str | None = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.baseDir.mkdir(parents=True, exist_ok=True)
        self.spriteBaseDir.mkdir(parents=True, exist_ok=True)


class Map(BaseModel):
    tileServer: TileServer = TileServer()
    zoom: int = 14
    pitch: int = 45
    lat: float = 12.0000
    lng: float = 79.8106
    bearing: float = 0
    style: str = "OpenFreeMap"
    opacity: float = 1
    attribution: str = ""
    status: list[str] = ["E", "F", "D"]
    defaultStatus: list[str] = ["E"]  # FIXME: should be str
    tagKeys: list[str] = ["source"]


class Measures(BaseModel):
    defaultStore: str | None = None


class BasketDefault(BaseModel):
    surveyor: str = "Default surveyor"
    equipment: str = "Default equipment"
    project: str = "Default project"
    status: str = "E"
    store: str | None = None


# class BasketOldDef(BaseModel):
#     base_dir: str


class Basket(BaseModel):
    base_dir: str = "/var/local/gisaf/baskets"
    default: BasketDefault = BasketDefault()


class Plot(BaseModel):
    maxDataSize: int = 10000


class Dashboard(BaseModel):
    base_source_url: str = "http://url.to.jupyter/lab/tree/"
    base_storage_dir: str = "/var/lib/share/gisaf/dashboard"
    base_storage_url: str = "/dashboard-attachment/"


class Widgets(BaseModel):
    footer: str = (
        """Generated by <span class='link' onclick="window.open('https://redmine.auroville.org.in/projects/gisaf/')">Gisaf</span>"""
    )


class Admin(BaseModel):
    basket: Basket = Basket()


class Attachments(BaseModel):
    base_dir: str = "/var/local/gisaf/attachments"


class Job(BaseModel):
    id: str
    func: str
    trigger: str
    minutes: int | None = 0
    seconds: int | None = 0


class Crs(BaseModel):
    """
    Handy definitions for crs-es
    """

    db: str
    geojson: str
    for_proj: str
    survey: str
    web_mercator: str


class Config(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="GISAF__",
        nested_model_default_partial_update=True,
        env_nested_delimiter="__",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        configs = [
            YamlConfigSettingsSource(settings_cls, yaml_file=cf) for cf in config_files
        ]
        return (
            env_settings,
            init_settings,
            file_secret_settings,
            *configs,
        )

    # def __init__(self, **kwargs):
    #     super().__init__(**kwargs)
    #     self.crs = {
    #         'db': f'epsg:{conf.srid}',
    #         'geojson': f'epsg:{conf.geojson_srid}',
    #         'for_proj': f'epsg:{conf.srid_for_proj}',
    #         'survey': f'epsg:{conf.raw_survey_srid}',
    #         'web_mercator': 'epsg:3857',
    #     }

    admin: Admin = Admin()
    attachments: Attachments = Attachments()
    # basket: BasketOldDef = BasketOldDef()
    # crs: Crs
    crypto: Crypto = Crypto()
    dashboard: Dashboard = Dashboard()
    db: DB = DB()
    # flask: Flask
    geo: Geo = Geo()
    gisaf: GisafConfig = GisafConfig()
    gisaf_live: GisafLive = GisafLive()
    jobs: list[Job] = []
    map: Map = Map()
    measures: Measures = Measures()
    ogcapi: OGCAPI = OGCAPI()
    plot: Plot = Plot()
    plugins: dict[str, dict[str, Any]] = {}
    survey: Survey = Survey()
    version: str = get_version()
    weather_station: dict[str, dict[str, Any]] = {}
    widgets: Widgets = Widgets()

    @property
    def crs(self) -> Crs:
        return Crs(
            db=f"epsg:{self.geo.srid}",
            geojson=f"epsg:{self.geo.srid}",
            for_proj=f"epsg:{self.geo.srid_for_proj}",
            survey=f"epsg:{self.geo.raw_survey.srid}",
            web_mercator="epsg:3857",
        )


conf = Config()

# def set_app_config(app) -> None:
#     raw_configs = []
#     with open(Path(__file__).parent / 'defaults.yml') as cf:
#         raw_configs.append(cf.read())
#     for cf_path in (
#         Path(Path.cwd().root) / 'etc' / 'gisaf' / ENV,
#         Path.home() / '.local' / 'gisaf' / ENV
#     ):
#         try:
#             with open(cf_path.with_suffix('.yml')) as cf:
#                 raw_configs.append(cf.read())
#         except FileNotFoundError:
#             pass

#     yaml_config = safe_load('\n'.join(raw_configs))

#     conf.app = yaml_config['app']
#     conf.postgres = yaml_config['postgres']
#     conf.storage = yaml_config['storage']
#     conf.map = yaml_config['map']
#     conf.security = yaml_config['security']
#     # create_dirs()


# def create_dirs():
#     """
#     Create the directories needed for a proper functioning of the app
#     """
#     ## Avoid circular imports
#     from treetrail.api_v1 import attachment_types
#     for type in attachment_types:
#         base_dir = Path(conf.storage['root_attachment_path']) / type
#         base_dir.mkdir(parents=True, exist_ok=True)
#     logger.info(f'Cache dir: {get_cache_dir()}')
#     get_cache_dir().mkdir(parents=True, exist_ok=True)


# def get_cache_dir() -> Path:
#     return Path(conf.storage['root_cache_path'])
